from typing import Optional, Any
from uuid import UUID

from sqlalchemy.engine import Dialect
import sqlalchemy.types as types


UUID_LENGTH = 36


class Uuid(types.TypeDecorator):
    """
    A UUID, stored as a 36-character string containing dashes.

    There is a built-in SQLAlchemy type representing a UUID (which uses built-in database types for UUIDs, if they
    exist), but it falls back to 32-character strings (dashes are removed) if there is no such type (like with
    MySQL, MariaDB or SQLite).

    This makes working directly in the database (which often is a requirement) complicated, so we've decided to
    implement our own type for that.
    """

    impl = types.TypeEngine
    cache_ok = True

    def load_dialect_impl(self, dialect: Dialect) -> types.TypeEngine[Any]:
        return types.String(UUID_LENGTH)

    def process_bind_param(self, value: str | UUID, dialect: Dialect) -> Optional[str]:
        if isinstance(value, str):
            # Manually create UUID from string, to raise an error if the string is malformed
            UUID(value)

        return str(value) if value else None

    def process_result_value(
        self, value: Optional[str], dialect: Dialect
    ) -> Optional[UUID]:
        return UUID(value) if value else None
